#ifndef BFM_THREAD_H
#define BFM_THREAD_H


/*
 *Thread model - threads created EXTERNAL to Bagel/BFM.
 *
 *             - Number of threads passed in.
 *
 *             - Threads register themselves with the bagel (get threadid)
 *
 *             - Boss thread 
 *               - does the alloc's, table inits etc...
 *
 *             - All threads call all numerical routines.
 *               - linalg Mprec etc... all to a per thread chunk of work
 *                 & contain barriers()
 *
 *             - CG contains allocs. These should be replaced with AllocBcast.
 *
 *             - Simply need a barrier & gsum across threads & bcast
 *
 *      Thus:
 *           Setup is single thread
 *           Fork and join is done at the ENTIRE CG level.
 *           Registration is tricky. Maps pthread_self to a work unit.
 */

void SpiL2AtomicInit(void);

class ThreadModelSingle {
 public:
  virtual void thread_init      (int nthr) { 
     if (nthr != 1) exit(0); 
     nthread=nthr; 
  };
  virtual int  thread_barrier   (void) { return 0; };
  virtual void thread_sum       (double &val, int me) { return; };
  virtual void thread_csum   (uint64_t &val,int me) {return;};
  virtual void  *thread_bcast(int me,void * val) { return val; };

  void thread_work_nobarrier(int nwork, int me, int & mywork, int & myoff){
    int basework = nwork/nthread;
    int backfill = nthread-(nwork%nthread);
    mywork = (nwork+me)/nthread;
    myoff  = basework * me;
    if ( me > backfill ) 
      myoff+= (me-backfill);

  }

  void thread_work(int nwork, int &me, int & mywork, int & myoff){
    me     = thread_barrier();
    thread_work_nobarrier(nwork,me,mywork,myoff);
  }

  int nthread;
};

#ifdef THREAD_MODEL_SINGLE

typedef ThreadModelSingle ThreadModel;

#endif

#ifdef THREAD_MODEL_PTHREAD

#include <pthread.h>
class ThreadModelPthread : public ThreadModelSingle {
 public:
  virtual void thread_init      (int nthread);
  virtual int  thread_barrier(void) ;//unique number between 0 and nthread-1
  virtual void thread_sum    (double &val,int me) ;
  virtual void thread_csum   (uint64_t &val,int me) ;
  virtual void  *thread_bcast(int me,void * val);
 protected:
  pthread_mutex_t barrier_lock;
  volatile unsigned char   barrier_step;
  volatile int             barrier_threads;
 public:
  double * sum_array;
  void   * bcast_ptr;
};
typedef ThreadModelPthread ThreadModel;

#endif

#ifdef THREAD_MODEL_SPI
#include <spi/include/l2/atomic.h>
//#include <spi/include/l2/barrier.h>

#define L1D_CACHE_LINE_SIZE 64

typedef struct {
  volatile __attribute__((aligned(L1D_CACHE_LINE_SIZE)))
    uint64_t start;  /*!< Thread count at start of current round. */
  volatile __attribute__((aligned(L1D_CACHE_LINE_SIZE)))
    uint64_t count;  /*!< Current thread count. */
} L2_Barrier_t;


#define MAX_HW_THREADS (512)
class ThreadModelSPI : public ThreadModelSingle {
 public:
  virtual void thread_init      (int nthread);
  virtual int  thread_barrier   (void) ;
  virtual void thread_sum       (double &val, int me) ;
  virtual void thread_csum   (uint64_t &val,int me) ;
  virtual void  *thread_bcast(int me,void * val);
  void thread_init_noalloc      (int nthread);

  //SPI specific ; prevent work migration between threads
  int  bgq_hwid(void);
  void bgq_hwid_register(int hwid);

  int ID_map[MAX_HW_THREADS];
  int ID_map_initialized;


 protected:
 public:
  double * sum_array;
  void   * bcast_ptr;
};
typedef ThreadModelSPI ThreadModel;
#endif

void SPI_PhysicalThreadMapping(int yesno);

#endif
